"""
Pepsin configuration handler
"""
from datetime import datetime
from typing import Any, Dict, List, Union

from pepsin.utils import check_file_exists, update_file
from pepsin.yml import YAMLConfig


def get_project_name(**options) -> str:
    """
    Returns project name, if project name is
    given use that name otherwise use default name `package`
    Args:
        **options: Dict | CLI arguments and CLI prompt data

    Returns: str | Name of the project

    """
    kw_proj_name: str = options.get("name", None)
    if kw_proj_name:
        return kw_proj_name

    return "package"


def get_or_empty_str(attr: Any) -> Union[Any, str]:
    """
    Mainly for the Pepsin config yaml file,
    PyYaml uses `null` for `None` or `empty` values,
    This function returns empty string if anything is null
    Args:
        attr: Any

    Returns: str | Any
    """
    return "" if not attr else attr


class PepsinConfig:
    """
    Pepsin Config Blueprint and handler
    1. Reads pepsin config / `pepsin.yaml` file
    2. Updates pepsin config
    3. Updates libraries
    4. Removes libraries
    """

    __slots__ = [
        "name",
        "venv",
        "author",
        "license",
        "libraries",
        "scripts",
        "__conf",
    ]

    def __init__(self):
        self.libraries = []
        self.__conf = YAMLConfig("pepsin.yaml")
        self.venv = "venv"
        self.set_config()

    def read_config(self) -> Dict:
        """
        Reads pepsin yaml config and returns dictionary
        Returns: Dict | Reads from yaml file and returns dictionary

        """
        return self.__conf.get_config()

    def set_config(self):
        """
        Sets instance config attributes after reading the
        config file
        Returns: None

        """
        conf = self.read_config()
        for slot in self.get_slots():
            setattr(self, slot, conf.get(slot))
        if not self.libraries:
            self.libraries = []

    def reload(self):
        """
        Reloads config if config has been updated
        Returns:

        """
        self.set_config()

    def get_slots(self):
        """
        Returns: List of slot excluding conf
        """
        return [slot for slot in self.__slots__ if slot != "__conf"]

    @staticmethod
    def config_exists() -> bool:
        """
        Checks if config file exists in the working directory
        Returns: bool
        """
        return check_file_exists("pepsin.yaml")

    def update(self, **kwargs):
        """
        Updates configuration and saves them in the yaml config
        Args:
            **kwargs: Dict | Values to update and save

        Returns:

        """
        libs = kwargs.get("libraries", [])
        # Update libraries
        self.update_libraries(libs)
        for key in kwargs:
            if key != "libraries" and key in self.get_slots():
                setattr(self, key, kwargs.get(key))
        # Update configuration and save to yaml
        self.__conf.append(**self.format_config())
        self.__conf.save()

    def update_libraries(self, libs: List[str]):
        """
        Appends / modified libraries in the config
        If one library already exists in the config, it will pass
        but if the library with different version exists it will be
        replaced with the new one
        Args:
            libs: List[str] | List of libraries

        Returns:

        """
        if not libs:
            return
        for lib in libs:
            # Find similar libraries and remove them
            # 'lib==2.3.4' replace with 'lib=2.3.5'
            # Remove 'lib==2.3.4'
            # Add 'lib==2.3.5'
            # Split using '==' and match first 2 strings
            for conf_lib in self.libraries:
                if lib.split("==")[0] in conf_lib.split("==")[0]:
                    index = self.libraries.index(conf_lib)
                    del self.libraries[index]

            self.libraries.append(lib)

    def remove_libraries(self, libs: List[str]):
        """
        Remove libraries from the config
        Args:
            libs: List[str] List of libraries

        Returns:

        """
        for lib in libs:
            if lib in self.libraries:
                self.libraries.remove(lib)
        # Update configuration and save to yaml
        self.__conf.append(**self.format_config())
        self.__conf.save()

    def format_config(self) -> Dict:
        """
        Formats configuration to dump into yaml file
        Returns: Dict | Dictionary of configurations
        """
        return {
            key: get_or_empty_str(getattr(self, key))
            for key in self.get_slots()
        }

    def initialize_config(self, **kwargs):
        """
        Initialize project configuration if there is no config
        Args:
            kwargs: Additional options
        """
        if not self.config_exists():
            self.update(**kwargs)


def handle_failed_libs(
    failed: List[str], action_msg: str = "Module Installation Failed"
):
    """
    Creates / Updates log file
    Args:
        failed: List[str] List of failed libraries
        action_msg: str | Message to put in the log file

    Returns:

    """
    ftext = (
        f"# {action_msg}" f' {datetime.now().strftime("%d %B %Y | %H:%M:%S")}'
    )
    failed.insert(0, ftext)
    update_file("pepsin.failed.log", "\n".join(failed))



"""
This module handles library installation
"""
from argparse import ArgumentParser

from pepsin.base import Base
from pepsin.config import PepsinConfig
from pepsin.pyhandler import PyHandler
from pepsin.utils import get_default


class Command(Base):
    """
    Install command class
    """

    short_description = "Install library"
    help = """Install a particular or multiple libraries
`$pepsin install <library>`
Example: `$pepsin install django`
"""

    def add_argument(self, parser: ArgumentParser):
        parser.add_argument(
            "libraries_to_install",
            nargs="*",
            metavar="libs",
            type=str,
            help="Libraries to install",
            default="",
        )

        parser.add_argument(
            "-r",
            type=str,
            metavar="requirement.txt",
            help=(
                "Install from a text file, similar to pip install -r"
                " requirements.txt"
            ),
        )

    def execute(self):
        """
        Installs library
        """
        # pepsin Config instance
        config = PepsinConfig()
        # If venv and no config is initialized then create venv and config
        config.initialize_config()
        py_handler = PyHandler(config)
        libs = get_default(self.command_data.get("libraries_to_install"), [])
        requirement = self.command_data.get("r")
        if not libs and not requirement:
            libs = config.libraries
        installed = py_handler.install_libraries(
            libs, requirements=requirement
        )[0]
        config.update(libraries=installed)



"""
Pepsin Errors
"""


class InvalidCommandError(Exception):
    """
    Exception class indicating a problem while executing a command.
    """

    def __init__(
        self,
        *args,
        return_code=1,
    ):
        self.return_code = return_code
        super().__init__(*args)


class TemplateDoesNotExistError(Exception):
    """
    Template Non Existence Error
    """



from test.utils import command_path, get_installed_libs_in_venv

import pytest

from pepsin.commands.install import Command as InstallCommand
from pepsin.commands.uninstall import Command as UninstallCommand
from pepsin.config import PepsinConfig


@pytest.fixture
def install_command():
    return InstallCommand()


@pytest.fixture
def uninstall_command():
    return UninstallCommand()


def test_uninstall_command(uninstall_command, install_command):
    conf = PepsinConfig()
    conf.update(venv="test_uninstall_venv")
    install_command.run(["pepsin", "install", "flask"])
    conf = PepsinConfig()
    assert conf.libraries == ["flask"]
    libraries = get_installed_libs_in_venv("test_uninstall_venv")
    assert "flask" in libraries
    uninstall_command.run(["pepsin", "uninstall", "flask"])
    conf = PepsinConfig()
    assert conf.libraries == []
    libraries = get_installed_libs_in_venv("test_uninstall_venv")
    assert "flask" not in libraries



"""
PIP Command module
"""
from argparse import ArgumentParser

from pepsin.base import Base
from pepsin.pyhandler import PyHandler


class Command(Base):
    """
    Executes Pip command
    Examples:
        `pip --h` now can be written as
        `pepsin pip --h`
    """

    short_description = "Execute pip commands"
    help = """Executes pip commands
        EG: `pepsin pip help`
        """

    def add_argument(self, parser: ArgumentParser):
        parser.add_argument(
            "command",
            metavar="<command>",
            nargs="*",
            help="Pip command, IE: freeze",
        )

    def execute(self):
        py_handler = PyHandler(skip_venv=True)
        if self.command_data.get("command"):
            py_handler.pip_execute(*self.argv)



from pepsin.error import TemplateDoesNotExistError
from pepsin.template import Template


def test_template_does_not_exist():
    template = Template(template_name="TEST")

    try:
        template.read()

    except TemplateDoesNotExistError as error:
        assert isinstance(error, TemplateDoesNotExistError)



"""
Ref: https://gist.github.com/nicodebo/b68034445da9be238e706a2d8166cf79
"""

import itertools
import sys
import threading
import time
from enum import Enum


class Sequence(Enum):
    """Enumeration of spinner sequence
    Ref: https://stackoverflow.com/questions/2685435/cooler-ascii-spinners
    """

    BASIC = ["-", "/", "|", "\\"]
    ARROW = ["←", "↖", "↑", "↗", "→", "↘", "↓", "↙"]
    VERT_BAR = ["▁", "▃", "▄", "▅", "▆", "▇", "█", "▇", "▆", "▅", "▄", "▃"]
    HORIZ_BAR = [
        "▉",
        "▊",
        "▋",
        "▌",
        "▍",
        "▎",
        "▏",
        "▎",
        "▍",
        "▌",
        "▋",
        "▊",
        "▉",
    ]
    SPIN_RECT = ["▖", "▘", "▝", "▗"]
    ELAST_BAR = ["▌", "▀", "▐▄"]
    TETRIS = ["┤", "┘", "┴", "└", "├", "┌", "┬", "┐"]
    TRIANGLE = ["◢", "◣", "◤", "◥"]
    SQUARE_QRT = ["◰", "◳", "◲", "◱"]
    CIRCLE_QRT = ["◴", "◷", "◶", "◵"]
    CIRCLE_HLF = ["◐", "◓", "◑", "◒"]
    BALLOON = [".", "o", "O", "@", "*"]
    BLINK = ["◡◡", "⊙⊙", "◠◠"]
    TURN = ["◜ ", " ◝", " ◞", "◟ "]
    LOSANGE = ["◇", "◈", "◆"]
    BRAILLE = ["⣾", "⣽", "⣻", "⢿", "⡿", "⣟", "⣯", "⣷"]

    # def describe(self):
    #     return self.name, self.value

    @classmethod
    def default_sequence(cls):
        """
        Returns default sequence
        """
        return Sequence.HORIZ_BAR


class Spinner:
    """A shell spinner"""

    def __init__(
        self, message="", interval=0.25, sequence="HORIZ_BAR", offset=1
    ):
        self.stop_running = threading.Event()
        self.spin_thread = threading.Thread(target=self.init_spin)
        self.interval = interval  # speed rotation
        self.message = message  # spinner text
        self.offset = offset  # number of space to pad on the left
        self.spinner_cycle = itertools.cycle(self.set_spinner_seq(sequence))

    @staticmethod
    def set_spinner_seq(sequence):
        """
        Sets spinner sequence from the Enum
        """
        seq_list = [name for name, members in Sequence.__members__.items()]
        if sequence.upper() in seq_list:
            seq = Sequence[sequence.upper()].value
        else:
            seq = Sequence.default_sequence().value
        return seq

    def start(self):
        """
        Starts spinning action
        """
        self.spin_thread.start()

    def stop(self):
        """
        Stops spinning action
        """
        self.stop_running.set()
        self.spin_thread.join()
        # make sure to clear the line in case printing something shorter after
        sys.stdout.write("\033[K")

    def init_spin(self):
        """
        Initialize Spinner spin
        """
        while not self.stop_running.is_set():
            if not self.message:
                cur_mess = next(self.spinner_cycle)
            else:
                cur_mess = f"{next(self.spinner_cycle)} {self.message}"
            cur_mess = cur_mess.rjust(len(cur_mess) + self.offset, " ")
            sys.stdout.write(cur_mess)
            sys.stdout.flush()
            time.sleep(self.interval)
            sys.stdout.write("\b" * len(cur_mess))



"""
Base class for writing commands, commands which will be executed
through the cli
"""
import sys
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from typing import Dict

from pepsin.base_io import IOBase, PromptHandler
from pepsin.error import InvalidCommandError
from pepsin.template import TemplateList
from pepsin.version import get_version


class Base(IOBase, ABC):
    """
    The base class which all commands will inherit
    and it contains the signature methods and attributes

    Attributes:
        1. ``help`` long description of the command
        for a certain command.

        2. ``short_description``
        short possible description to describe a command
        or signify the uses.

        3. ``command_data`` Contains parsed arguments and cli prompt answers
        formatted correctly.

    Methods:
        1. ``add_argument`` Adds custom argument to the parser,
        that will be later converted into dictionary
        Examples
            ```
            def add_argument(self, parser):
                parser.add_argument('--option_1', action='store_true', help='Help Text')
            ```

        2. ``add_prompt`` Adds interactive input and output
        Examples
            ```
            def add_prompt(self, prompt_handler):
                parser.add_prompts(
                    Input(),
                    Output(),
                )
            ```

        3. ``add_template`` Adds templates that will be generated
        after the cli command gets executed
        Examples
            ```
                def add_templates(self, template_list: TemplateList):
                    template_list.add_template(
                        Template(template_name=".gitignore"),
                        Template(
                            template_name="Readme.MD",
                            context={"name": self.command_data.get("name")},
                        ),
                    )
            ```

        4. ``execute`` method is the abstract method
        that needs to be implemented in the subclass, every
        command must have an ``execute`` method
        as business logic will be inside this method
    """

    help = ""
    short_description = ""

    def __init__(self):
        self.prompt_handler: PromptHandler = PromptHandler()
        self.templates: TemplateList = TemplateList()
        self.command_data = {}
        self.command_args = None
        self.argv = []
        super().__init__()

    def add_prompts(self, handler: PromptHandler, **options: Dict):
        """

        Args:
            handler: PromptHandler | PromptHandler Class instance
            **options: Dict | Dictionary containing cli arguments
                       parsed by argparse

        Returns: None

        """

    def add_argument(self, parser: ArgumentParser):
        """
        Subclass method, that will add required arguments
        for the dedicated command
        Args:
            parser: ArgumentParser

        Returns:

        """

    def add_templates(self, template_list: TemplateList):
        """
        Subclass method to add required templates for the dedicated
        command
        Args:
            template_list:

        Returns:

        """

    def format_help(self, command: str) -> str:
        """
        Returns help text
        Args:
            command: str | Name of the command that will be executed

        Returns: str | Formatted help text

        """
        return self.handle_parser(command).format_help()

    def format_arguments(self, **kwargs) -> Dict:
        """
        Formats cli arguments, mainly for the subclass to
        change or edit any options / cli arguments
        Args:
            **kwargs:

        Returns: Dict | Formatted CLI Argument / Options

        """
        return kwargs

    def handle_parser(self, command: str) -> ArgumentParser:
        """
        Creates Argument Parser Class Object, to parse cli arguments
        Args:
            command: str | Name of the command that will be executed

        Returns: ArgumentParser instance

        """
        parser = ArgumentParser(
            prog=f"Pepsin | {command}", description=self.help
        )
        parser.add_argument(
            "--version",
            action="version",
            version=get_version(),
        )
        parser.add_argument("--no-input", action="store_true")
        # Subclass arguments will be added via this method
        self.add_argument(parser)
        return parser

    def handle_input(self, **options):
        """
        This method handles cli inputs, and enquire from the user,
        if the subclass overrides ``add_prompts`` method, containing
        some CLIIoBase Subclass instances, it will call the ``prompt`` method
        of the PromptHandler Instance and take input data from the user and show
        given output
        Args:
            **options: Dict | Dictionary containing cli arguments
                       parsed by argparse

        Returns:

        """
        if options.get("no_input"):
            return
        self.prompt_handler.update_options(options)
        self.add_prompts(self.prompt_handler, **options)
        self.prompt_handler.prompt()

    def handle_template(self):
        """
        Handles subclass added templates
        Returns:

        """
        self.add_templates(self.templates)
        self.templates.save()

    @abstractmethod
    def execute(self):
        """
        Subclass/Command business logic,
        main method where the command's workflow will be declared
        """

    def run(self, argv) -> None:
        """
        Runner method for command class
        """
        # Get the parser to get cli arguments and process
        self.argv = argv[2:]
        parser = self.handle_parser(command=argv[1])
        options = vars(parser.parse_args(argv[2:]))
        args = options.pop("args", ())
        # Handle inputs from the cli
        self.handle_input(**options)
        # Format all inputs
        options.update(**self.prompt_handler.get_answers())
        options = self.format_arguments(**options)
        options.pop("no_input")
        self.command_data = options
        self.command_args = args
        try:
            self.execute()
            self.handle_template()
        except InvalidCommandError as error:
            self.error(f"{error.__class__.__name__}: {error}")
            sys.exit(error.return_code)



import os
import shutil
from pathlib import Path

import pytest

from pepsin.utils import OSEnum, check_dir_exists, get_os


def make_multiple_inputs(inputs: list):
    """provides a function to call for every input requested."""

    def next_input(_):
        """provides the first item in the list."""
        return inputs.pop(0)

    return next_input


def safe_remove_dir(dir_path):
    try:
        shutil.rmtree(dir_path)
    except FileNotFoundError:
        pass


def safe_remove_file(file_path):
    try:
        os.remove(file_path)
    except FileNotFoundError:
        pass


@pytest.fixture
def temp_path():
    path = Path(__file__).resolve().parent / "temp"
    os.chdir(path=path)
    return path


@pytest.fixture(autouse=True)
def set_subprocess(monkeypatch):
    def check_call(*args, **kwargs):
        pass

    monkeypatch.setattr("subprocess.check_call", check_call)
    yield


@pytest.fixture(autouse=True)
def command_path():
    path = Path(__file__).resolve().parent / "temp"
    os.chdir(path=path)
    try:
        if check_dir_exists(path, "__TEST__"):
            safe_remove_dir("__TEST__")
        os.mkdir("__TEST__")
    except PermissionError:
        pass
    os.chdir(path=path / "__TEST__")
    yield
    os.chdir(path=path)
    safe_remove_dir("__TEST__")
    safe_remove_dir("venv")
    safe_remove_file("pepsin.yaml")
    safe_remove_file("pepsin.failed.log")


def get_installed_libs_in_venv(venv_name):
    if get_os() == OSEnum.WIN:
        dirs = os.listdir(os.path.join(venv_name, "lib", "site-packages"))
    else:
        py_dir = os.listdir(
            os.path.join(
                venv_name,
                "lib",
            )
        )
        dirs = os.listdir(
            os.path.join(venv_name, "lib", py_dir[0], "site-packages")
        )

    return dirs



"""
This module handles library Upgrade
"""
from argparse import ArgumentParser

from pepsin.base import Base
from pepsin.config import PepsinConfig
from pepsin.pyhandler import PyHandler
from pepsin.utils import get_default


class Command(Base):
    """
    Upgrade command class
    """

    short_description = "Update or upgrade library"
    help = """Upgrades a particular or multiple libraries
    `$pepsin upgrade <library>`
    Example: `$pepsin upgrade django`
    """

    def add_argument(self, parser: ArgumentParser):
        parser.add_argument(
            "-r",
            type=str,
            metavar="<requirement.txt>",
            help="Upgrade from a text file",
        )

        parser.add_argument(
            "libs",
            nargs="*",
            metavar="<Libraries>",
            type=str,
            help="Libraries to upgrade",
            default="",
        )

    def execute(self):
        """
        Installs library
        """
        # pepsin Config instance
        config = PepsinConfig()
        # If venv and no config is initialized then create venv and config
        config.initialize_config()
        py_handler = PyHandler(config)
        libs = get_default(self.command_data.get("libs"), [])
        requirement = self.command_data.get("r")
        if not libs and not requirement:
            libs = config.libraries
        installed = py_handler.upgrade_libraries(
            libs, requirements=requirement
        )[0]
        config.update(libraries=installed)



import os
from test.utils import command_path, make_multiple_inputs

import pytest

from pepsin.commands.init import Command as InitCommand
from pepsin.main import CLI


@pytest.fixture
def init_command():
    return InitCommand()


def test_init_with_option_args(init_command, monkeypatch):
    # Set input parameter
    monkeypatch.setattr(
        "builtins.input",
        make_multiple_inputs(
            ["testproject", "mit", "author", "name@name.com"]
        ),
    )
    # Run the command
    init_command.run(["pepsin", "init", "--venv=testvenv"])
    # Check directories
    assert os.path.isdir("testvenv")
    assert os.path.isdir("testproject")
    assert os.path.isfile("pepsin.yaml")
    assert os.path.isfile("testproject/main.py")
    assert os.path.isfile("Readme.MD")
    assert os.path.isfile(".gitignore")


def test_init_with_predefined_files_no_input(init_command, monkeypatch):
    cli = CLI(
        ["pepsin", "init", "testproject", "--venv=testvenv", "--no-input"]
    )
    # Set input parameter
    os.mkdir("testproject")
    with open("testproject/main.py", "w") as file:
        file.write("print('Hello World')")
        file.close()
    # Run the command
    cli.execute()
    # Check directories
    assert os.path.isdir("testvenv")
    assert os.path.isdir("testproject")
    assert os.path.isfile("pepsin.yaml")
    assert os.path.isfile("testproject/main.py")
    assert os.path.isfile("Readme.MD")
    assert os.path.isfile(".gitignore")
    # Read if file has not been updated
    with open("testproject/main.py", "r") as file:
        assert "Hello World" in file.read()
        file.close()


def test_init_with_predefined_files(monkeypatch):
    cli = CLI(["pepsin", "init", "testproject", "--venv=testvenv"])
    # Set input parameter
    monkeypatch.setattr(
        "builtins.input",
        make_multiple_inputs(["mit", "author", "name@name.com"]),
    )

    os.mkdir("testproject")
    with open("testproject/main.py", "w") as file:
        file.write("print('Hello World')")
        file.close()
    # Run the command
    cli.execute()
    # Check directories
    assert os.path.isdir("testvenv")
    assert os.path.isdir("testproject")
    assert os.path.isfile("pepsin.yaml")
    assert os.path.isfile("Readme.MD")
    assert os.path.isfile(".gitignore")
    assert os.path.isfile("testproject/main.py")
    # Read if file has not been updated
    with open("testproject/main.py", "r") as file:
        assert "Hello World" in file.read()
        file.close()



"""
This module handles python and pip execution
"""
import os
import subprocess
import sys
from typing import List, Optional
from urllib import request
from urllib.error import HTTPError, URLError

from pepsin.base_io import OutputWrapper
from pepsin.config import PepsinConfig, handle_failed_libs
from pepsin.const import PIP_DL_LINK
from pepsin.utils import (
    OSEnum,
    check_dir_exists,
    check_file_exists,
    get_default,
    get_os,
    read_file,
    write_file,
)


class PyHandler:
    """
    Handles python and pip package installation, management
    """

    def __init__(
        self,
        pepsin_config: PepsinConfig = None,
        stdout: OutputWrapper = None,
        stderr: OutputWrapper = None,
        skip_venv: bool = False,
    ):
        """
        Save env variable and set env variable to run code
        Using venv
        Args:
            pepsin_config: pepsinConfig Instance
            stdout: OutputWrapper(sys.stdout) Instance
            stderr: OutputWrapper(sys.stderr) Instance
            skip_venv: Skip initializing venv on init
        """
        # Read pepsin config to get venv
        self.output = stdout if stdout else OutputWrapper(sys.stdout)
        self.error = stderr if stderr else OutputWrapper(sys.stderr)
        self.executable = "python" if get_os() == OSEnum.WIN else "python3"
        self.pip_exec = "pip" if get_os() == OSEnum.WIN else "pip3"
        self.pepsin_config = pepsin_config if pepsin_config else PepsinConfig()
        self.env = os.environ.copy()

        self.venv = self.pepsin_config.venv
        # Create virtualenv if not exist
        if (not self.venv and not skip_venv) or (
            self.venv and not check_dir_exists(self.venv)
        ):
            self.venv = get_default(self.venv, "venv")
            self.pepsin_config.update(venv=self.venv)
            self.init_venv(self.venv)

        self.set_env()

    @staticmethod
    def __join_env_path(script_dir):
        """
        Joins script path and Returns ENV Path
        Args:
            script_dir: Venv Script Path

        Returns: string
        """
        return os.pathsep.join(
            [script_dir] + os.environ.get("PATH", "").split(os.pathsep)
        )

    def set_env(self):
        """
        Sets up environment variable

        Returns: None
        """
        if self.venv:
            sys_os = get_os()
            script_loc = "bin"
            if sys_os == OSEnum.WIN:
                script_loc = "scripts"
            venv_dir = os.path.join(os.getcwd(), self.venv)
            script_dir = os.path.join(os.getcwd(), self.venv, script_loc)
            self.env["VIRTUAL_ENV"] = venv_dir
            self.env["PATH"] = self.__join_env_path(script_dir)
            self.executable = f"{script_dir}/python"
            self.pip_exec = f"{script_dir}/pip"

    def python_execute(self, *commands):
        """
        Executes python command
        Returns:
        """
        subprocess.check_call([self.executable, *commands], env=self.env)

    def init_venv(self, venv_dir: str):
        """
        Initializes virtualenv directory
        Returns:
        """
        self.python_execute("-m", "virtualenv", venv_dir)

    def pip_execute(self, *commands):
        """
        Executes pip command
        Returns:

        """
        subprocess.check_call([self.pip_exec, *commands], env=self.env)

    def pip_install(self, *packages):
        """
        Installs package using pip
        Args:
            *packages:

        Returns:

        """
        subprocess.check_call(
            [self.pip_exec, "install", *packages], env=self.env
        )

    def pip_upgrade(self, *packages):
        """
        Upgrade python packages using pip
        Args:
            *packages: List[str]

        Returns:

        """
        package_list: List[str] = list(packages)
        # This part of the code safely upgrades pip
        # Default upgrade can cause unusual pip behaviors
        has_pip = False
        if "pip" in package_list or "pip3" in package_list:
            has_pip = True
            to_remove = "pip3" if "pip3" in package_list else "pip"
            package_list.remove(to_remove)
        if has_pip:
            try:
                with request.urlopen(PIP_DL_LINK) as file:
                    write_file("get_pip.py", file.read().decode("utf-8"))
                    self.python_execute("get_pip")
            except (URLError, HTTPError):
                self.error.write("Unable to upgrade pip")
        subprocess.check_call(
            [self.pip_exec, "install", "--upgrade", *package_list],
            shell=True,
            env=self.env,
        )

    def __process_library(
        self, action: str, libs: Optional[List[str]] = None, requirements=""
    ) -> (List[str], List[str]):
        """
        Process library with action
        Args:
            action: Option["install" , "upgrade"]
            libs: List of libraries
            requirements: Requirement.txt or text requirement file
        Returns:
            Tuple: List of passed and failed libraries

        """
        passed = []
        failed = []
        to_install: List[str] = get_default(libs, [])
        if requirements:
            if not check_file_exists(requirements):
                self.error.write(f"{requirements} does not exist")
            else:
                to_install += read_file(requirements).split("\n")
                to_install = [each for each in to_install if "#" not in each]
        for lib in to_install:
            try:
                if action == "upgrade":
                    self.pip_upgrade(lib)
                else:
                    self.pip_install(lib)
                if lib not in ["pip", "pip3"]:
                    passed.append(lib)
            except subprocess.CalledProcessError:
                self.error.write(f"Unable to install {lib}")
                failed.append(lib)

        if failed:
            handle_failed_libs(failed)

        return passed, failed

    def install_libraries(
        self, libs: Optional[List[str]] = None, requirements=""
    ) -> (List[str], List[str]):
        """
        Installs multiple libraries
        Args:
            libs: List of libraries
            requirements : Requirement.txt or text requirement file
        Returns:
            tuple: List of passed and failed libraries
        """
        return self.__process_library("install", libs, requirements)

    def upgrade_libraries(
        self, libs: Optional[List[str]] = None, requirements=""
    ) -> (List[str], List[str]):
        """
        Upgrades multiple libraries
        Args:
            libs: List of libraries
            requirements : Requirement.txt or text requirement file
        Returns:
            tuple: List of passed and failed libraries
        """
        return self.__process_library("upgrade", libs, requirements)

    def uninstall_libraries(self, libs=None) -> tuple:
        """
        Args:
            libs: List of libraries to uninstall
        Returns:
            tuple: Tuple[List[str], List[str]] List of
                   libs that passed and failed to uninstall
        """
        passed = []
        failed = []
        to_uninstall = list(get_default(libs, []))
        for lib in to_uninstall:
            if lib in ["pip", "pip3"]:
                self.error.write("Unable to uninstall pip")
            else:
                try:
                    subprocess.check_call(
                        [self.pip_exec, "uninstall", lib, "-y"], env=self.env
                    )
                    passed.append(lib)
                except subprocess.CalledProcessError:
                    self.error.write(f"Unable to uninstall {lib}")
                    failed.append(lib)
        if failed:
            handle_failed_libs(failed, "Unable to uninstall")
        return passed, failed



from test.utils import safe_remove_file, temp_path

from pepsin.utils import (
    check_file_exists,
    get_default,
    read_file,
    update_file,
    write_file,
)


def test_get_default(temp_path):
    test_data = None
    data = get_default(test_data, "test")
    assert data == "test"


def test_read_file(temp_path):
    read = read_file("test.txt")
    assert read == ""


def test_update_file(temp_path):
    write_file("test_write.txt", "test")
    update_file("test_write.txt", "update")
    read = read_file("test_write.txt")
    assert "test" in read
    assert "update" in read
    safe_remove_file("test_write.txt")


def test_check_file_exists():
    assert not check_file_exists("dummy.txt")



import pytest

from pepsin.commands.pip import Command as PipCommand


@pytest.fixture
def pip_command():
    return PipCommand()


def test_pip_command(pip_command):
    pip_command.run(["pepsin", "pip", "help"])



"""
Install/Add Alternative
"""
from pepsin.commands.install import Command as InstallCommand


class Command(InstallCommand):
    """
    Install Command alternative
    """

    short_description = "Installs library"
    help = """Install a particular or multiple libraries
        `$pepsin i <library>`
        Example: `$pepsin i django`
        """



import os
from test.utils import command_path

import pytest

from pepsin.pyhandler import PyHandler
from pepsin.utils import write_file


@pytest.fixture
def py_handler():
    return PyHandler()


def test_py_handler_python_exec(py_handler):
    write_file("py_init.py", "print('pepsin')")
    py_handler.python_execute("py_init.py")


def test_py_handler_pip_command(py_handler):
    py_handler.pip_execute("--h")


def test_py_handler_skip_venv():
    py_handler = PyHandler(skip_venv=True)
    py_handler.pip_execute("--h")
    assert not os.path.isdir("venv")



import os
from test.utils import command_path, get_installed_libs_in_venv

import pytest

from pepsin.commands.install import Command
from pepsin.config import PepsinConfig
from pepsin.utils import OSEnum, check_dir_exists, check_file_exists, get_os


@pytest.fixture
def install_command():
    return Command()


def test_fail_requirement(install_command, monkeypatch):
    install_command.run(["pepsin", "install", "test2test"])
    assert check_file_exists("pepsin.failed.log")


def test_install_command_within_dir(install_command):
    conf = PepsinConfig()
    conf.update(libraries=["requests"], venv="install_venv")
    install_command.run(["pepsin", "install"])
    assert check_dir_exists("install_venv")
    dirs = get_installed_libs_in_venv("install_venv")
    assert "requests" in dirs



import os
import subprocess
from pathlib import Path

import pytest

from pepsin.const import COMMAND_NOT_FOUND_ERROR
from pepsin.main import CLI, main
from pepsin.utils import OSEnum, get_os
from pepsin.utils.spinner import Spinner
from pepsin.version import get_version


@pytest.fixture
def path():
    path = Path(__file__).resolve().parent / "temp"
    os.chdir(path=path)
    return path


def test_main_help(monkeypatch, capsys):
    monkeypatch.setattr("sys.argv", ["pepsin", "help"])
    main()
    assert "Available commands" in capsys.readouterr().out


def test_main_command_error(monkeypatch, capsys):
    monkeypatch.setattr("sys.argv", ["pepsin", "ins"])
    main()
    assert COMMAND_NOT_FOUND_ERROR in capsys.readouterr().err


def test_main_command_help(monkeypatch):
    cli = CLI(["pepsin", "init", "-h"])
    output = cli.help_text()
    assert "init" in output


def test_main_cli_get_arg(monkeypatch):
    cli = CLI(["pepsin", "init", "-h"])
    assert cli.get_arg(3) is None


def test_version(capsys):
    cli = CLI(["pepsin", "--version"])
    cli.execute()
    assert get_version() in capsys.readouterr().out


def test_command_help():
    cli = CLI(["__main__.py"])
    cli.execute()
    assert 1


def test_spinner():
    spinner = Spinner(sequence="TEST", message="Message")
    spinner.start()
    spinner.stop()
    assert 1


def test_default_spinner():
    spinner = Spinner()
    spinner.start()
    spinner.stop()
    assert 1


def check_all(lst=None):
    pass


def test_sys_platform():
    # Test Linux system
    platform_os = get_os("linux")
    assert platform_os == OSEnum.LINUX
    # Test windows system
    platform_os = get_os("darwin")
    assert platform_os == OSEnum.OSX
    # Test windows system
    platform_os = get_os("win32")
    assert platform_os == OSEnum.WIN



import subprocess
import time
from test.utils import command_path, set_subprocess

import pytest

from pepsin.commands.install import Command
from pepsin.config import PepsinConfig
from pepsin.utils import check_file_exists, write_file


@pytest.fixture
def install_command():
    return Command()


def test_install(install_command):
    install_command.run(
        ["pepsin", "install", "flask", "-r", "requirements.txt"]
    )
    conf = PepsinConfig()
    configs = conf.format_config()
    assert configs.get("libraries") == ["flask"]


def test_install_with_req(install_command):
    write_file("requirements.txt", "\n".join(["django", "flask"]))
    install_command.run(["pepsin", "install", "-r", "requirements.txt"])
    conf = PepsinConfig()
    assert conf.libraries == ["django", "flask"]


def test_install_without_req(install_command):
    conf = PepsinConfig()
    conf.update(libraries=["django", "flask"])
    install_command.run(["pepsin", "install"])
    assert conf.libraries == ["django", "flask"]



"""
Init Command
Usage:
    `pepsin init`

Generates a python boilerplate normal project along
with virtual environment.

`init` will prompt for 4 things such as
1. project_name
2. License
3. Author
4. Email

and will create a project folder under the name
`project_name`

Optional Parameters:

`pepsin init my_project`
Will create project without asking for project_name

`--venv=myVenv`
Custom venv directory

`--no-input`
Will not take any input
"""
import os
from argparse import ArgumentParser
from typing import Dict

from colorama import Fore

from pepsin.base import Base
from pepsin.base_io import Input, Output
from pepsin.config import PepsinConfig, get_project_name
from pepsin.pyhandler import PyHandler
from pepsin.template import Template, TemplateList

MAIN_FILE = ""
"""# Generated with pepsin

if __name__ == "__main__":
    print("Egg and spam")
"""


class Command(Base):
    """
    Init command class
    """

    short_description = "Initialize Virtual environment"
    help = """Initialize virtual environment, project data and file
    """

    def add_argument(self, parser: ArgumentParser):
        """
        Adds custom argument
        """
        parser.add_argument(
            "name",
            nargs="?",
            metavar="name",
            type=str,
            help="Name of Project",
            default="",
        )

        parser.add_argument(
            "--venv", help="Virtual Environment directory name", default="venv"
        )

    def add_prompts(self, handler, **options):
        """
        Handles prompt input
        """
        handler.add_prompts(
            Output(
                name="Welcome",
                title=f"Pepsin Setup Project\n{'-'*10}",
                color=Fore.CYAN,
            ),
            Input(
                name="name",
                title="Project Name",
                default="src",
                required=False,
                skip=True,
            ),
            Input(
                name="author", title="Author", default="Author", required=False
            ),
            Input(
                name="email",
                title="Email",
                default="email@email.com",
                required=False,
            ),
            Input(
                name="license", title="license", default="MIT", required=False
            ),
        )

    def format_arguments(self, **kwargs):
        """
        Formats command arguments and input
        """
        project_name = get_project_name(**kwargs)
        kwargs.update(name=project_name)
        return kwargs

    def handle_directory(self):
        """
        Handle directory creation
        """
        project_name = self.command_data.get("name")
        try:
            os.mkdir(project_name)
        except FileExistsError:
            pass

        with open(f"{project_name}/__init__.py", "w", encoding="utf-8"):
            pass
        working_dir = os.getcwd()

        if not os.path.exists(f"{working_dir}/{project_name}/main.py"):
            with open(
                f"{project_name}/main.py", "w", encoding="utf-8"
            ) as file:
                file.write(MAIN_FILE)

    def add_templates(self, template_list: TemplateList):
        """
        Adds custom template
        """
        template_list.add_template(
            Template(template_name=".gitignore"),
            Template(
                template_name="Readme.MD",
                context={"name": self.command_data.get("name")},
            ),
        )

    def execute(self):
        """
        Inherited execute method
        """
        conf = PepsinConfig()
        conf.update(**self.command_data)
        PyHandler(conf, self.stdout, self.stderr)
        # Installs virtualenv library
        self.handle_directory()
        self.output("\nProject initialization complete")



import os
from pathlib import Path
from test.utils import safe_remove_file

import pytest

from pepsin.config import PepsinConfig


@pytest.fixture
def path():
    path = Path(__file__).resolve().parent / "temp"
    os.chdir(path=path)
    return path


@pytest.fixture(autouse=True)
def run_around_tests():
    yield
    path = Path(__file__).resolve().parent / "temp"
    os.chdir(path=path)
    safe_remove_file("pepsin.yaml")


def test_yaml_config(path):
    conf = PepsinConfig()
    conf.update(name="Project")
    config = conf.format_config()
    assert config.get("name") == "Project"
    assert conf.config_exists()
    safe_remove_file("pepsin.yaml")


def test_yaml_libraries():
    conf = PepsinConfig()
    conf.update_libraries([])
    conf.update_libraries(["test==2.3.5"])
    assert "test==2.3.5" in conf.format_config().get("libraries")
    conf.update_libraries(["test==2.3.6"])
    assert "test==2.3.6" in conf.format_config().get("libraries")
    safe_remove_file("pepsin.yaml")


def test_init_config():
    conf = PepsinConfig()
    conf.initialize_config()
    assert conf.config_exists()



from collections import deque
from test.utils import make_multiple_inputs

import pytest

from pepsin.base_io import Input, IOBase, Output, PromptHandler


@pytest.fixture
def io_base():
    return IOBase()


def test_input_str(monkeypatch):
    """
    Test Input Type String
    """
    monkeypatch.setattr("builtins.input", lambda _: "test")
    string_input = Input(name="name", type=str, title="title")
    assert string_input.prompt_as_dict() == {"name": "test"}


def test_input_int(monkeypatch):
    """
    Test Input Type Integer
    """
    monkeypatch.setattr("builtins.input", lambda _: "1")
    int_input = Input(name="name", type=int, title="title")
    assert int_input.prompt_as_dict() == {"name": 1}


def test_input_boolean(monkeypatch):
    """
    Test Input Type Boolean
    """
    monkeypatch.setattr("builtins.input", lambda _: "y")
    int_input = Input(name="name", type=bool, title="title")
    assert int_input.prompt_as_dict() == {"name": True}


def test_input_options(monkeypatch):
    """
    Test Input Type Options
    """
    monkeypatch.setattr("builtins.input", lambda _: "2")
    int_input = Input(
        name="name", type=int, title="title", options=["option 1", "option 2"]
    )
    assert int_input.prompt_as_dict() == {"name": "option 2"}


def test_input_options_skip(monkeypatch):
    """
    Test Input Type Options
    """
    monkeypatch.setattr(
        "builtins.input",
        make_multiple_inputs(
            [
                "5",
                "1",
            ]
        ),
    )
    int_input = Input(
        name="name", type=int, title="title", options=["option 1", "option 2"]
    )
    assert int_input.prompt_as_dict() == {"name": "option 1"}


def test_input_str_skip(monkeypatch):
    """
    Test input skip while type string
    """
    monkeypatch.setattr("builtins.input", lambda _: "")
    string_input = Input(name="name", type=str, title="title", required=False)
    assert string_input.prompt_as_dict() == {"name": None}


def test_input_bool_skip(monkeypatch):
    """
    Test input skip while type boolean
    """
    monkeypatch.setattr("builtins.input", lambda _: "n")
    string_input = Input(name="name", type=bool, title="title", required=False)
    assert string_input.prompt_as_dict() == {"name": False}


def test_input_default(monkeypatch):
    """
    Test default data
    """
    monkeypatch.setattr("builtins.input", lambda _: "")
    string_input = Input(
        name="name", type=str, default="value", title="title", required=False
    )
    assert string_input.prompt_as_dict() == {"name": "value"}


def test_output(capsys):
    """
    Test output
    """
    output = Output(title="Title", name="Name")
    assert output.prompt_as_dict() == {}
    assert "Title" in capsys.readouterr().out


def test_input_handler(monkeypatch):
    monkeypatch.setattr("builtins.input", lambda _: "test")
    handler = PromptHandler(
        Input(
            name="name",
            type=str,
            default="value",
            title="title",
            required=False,
        )
    )
    handler.prompt()
    assert handler.get("name") == "test"


def test_io_base(io_base):
    io_base.output("Output")
    io_base.error("Error")
    io_base.output_flush()
    io_base.error_flush()
    assert 1



"""
Handles Yaml Config and converts from yaml to dictionary or
dictionary to yaml
"""
import copy
import os
from typing import Any

import yaml


def dict_to_yaml(_dict: dict, filename: str):
    """
    Converts dictionary to yaml
    Args:
        _dict: Python dictionary or hash map
        filename: Name of yaml file
    Returns:

    """
    with open(filename, "w", encoding="utf-8") as file:
        yaml.dump(_dict, file, sort_keys=False)


def yaml_to_dict(filename: str) -> dict:
    """
    Converts yaml  to  dictionary
    Args:
        filename: Name of yaml file

    """
    with open(filename, "r", encoding="utf-8") as file:
        return yaml.load(file, yaml.Loader)


class YAMLConfig:
    """
    Yaml Model controller
    1. Reads yaml file
    2. Saves dictionary to yaml
    3. Reads and stores yaml as dictionary
    4. Appends data to the yaml
    5. Removes data from the yaml
    """

    def __init__(self, filename: str = "temp.yaml", **kwargs):
        self.__config = {}
        self.__filename = ""
        self.__filename = filename
        self.__read_yaml()
        self.append(**kwargs)

    def __read_yaml(self):
        """
        Reads yaml and converts to dictionary and stores it to
        self.__config
        Returns: None
        """
        if os.path.isfile(self.__filename):
            yaml_data = yaml_to_dict(self.__filename)
            self.append(**yaml_data)

    def read_from_yaml(self) -> dict:
        """
        Reads directly from the yaml
        Returns: Yaml Config dictionary
        """
        if os.path.isfile(self.__filename):
            return yaml_to_dict(self.__filename)
        return self.__config

    def append(self, **kwargs):
        """
        Inserts data in the config dictionary
        Returns: None
        """
        self.__config.update(**kwargs)
        for key, val in self.__config.items():
            setattr(self, key, val)

    def remove(self, key) -> Any:
        """
        Removes a given key from the config
        Returns: Any | Removed element from the config
        """
        return self.__config.pop(key)

    def get(self, key) -> Any:
        """
        Returns: Any | Any given element using the key
        """
        return self.__config.setdefault(key, None)

    def get_config(self) -> dict:
        """
        Returns: Dict | returns the copy of the config
        """
        return copy.deepcopy(self.__config)

    def get_filename(self) -> str:
        """
        Returns: str | Yaml Filename
        """
        return self.__filename

    def save(self):
        """
        Saves updated config to the yaml file
        Returns: Dict | self.__config - Yaml config file
        """
        dict_to_yaml(self.__config, self.__filename)
        return self.get_config()



"""
Similar to install
"""
from pepsin.commands.install import Command as InstallCommand


class Command(InstallCommand):
    """
    Install/add command class
    """

    short_description = "Install library"
    help = """Adds/Installs a particular or multiple libraries
    `$pepsin add <library>`
    Example: `$pepsin add django`
    """



"""
Contains Base utility that is required to run the pepsin CLI Class
"""
import enum
import functools
import os
import pkgutil
from importlib import import_module
from pathlib import Path
from sys import platform

from pepsin.base import Base

ROOT = Path(__file__).resolve().parent


def find_commands():
    """
    Finds commands located in the commands directory
    """
    function_directory = os.path.join(ROOT, "../commands")

    return [
        name
        for _, name, is_pkg in pkgutil.iter_modules([function_directory])
        if not is_pkg and not name.startswith("_")
    ]


def load_command_class(name) -> Base:
    """
    Every Command module has a Command class, which will be imported
    :return Command Class
    """
    module = import_module(f"pepsin.commands.{name}")
    return module.Command()


@functools.lru_cache(maxsize=None)
def get_commands():
    """
    Set of commands that will be returned
    :return: set of commands
    """
    commands = set(find_commands())
    return commands


def check_file_exists(file):
    """
    Checks if file exists in working directory
    """
    return os.path.isfile(os.path.join(os.getcwd(), file))


def get_default(target, replace):
    """
    Checks if something is null or empty otherwise return the replacement
    """
    return replace if not target else target


def read_file(file):
    """
    Safely read File and return text
    """
    try:
        with open(file, "r", encoding="utf-8") as __file:
            return __file.read()
    except FileNotFoundError:
        return ""


def write_file(file, text):
    """
    Safely write file
    """
    with open(file, "w", encoding="utf-8") as __file:
        return __file.write(text)


def update_file(file, text):
    """
    Safely updates a file
    """
    read_text = read_file(file)
    write_file(file, read_text + "\n" + text)


def check_dir_exists(*paths):
    """
    Checks if a particular directory exists or not
    Args:
        *paths: Paths to join and check for directory existence

    Returns:

    """
    return os.path.isdir(os.path.join(*paths))


class OSEnum(enum.Enum):
    """
    OS Enum
    """

    LINUX = "linux"
    WIN = "win"
    OSX = "darwin"


def get_os(plt=platform) -> OSEnum:
    """
    Returns System OS Platform name
    """
    __os = OSEnum.LINUX
    if plt in ["linux", "linux2"]:
        __os = OSEnum.LINUX
    elif plt == "darwin":
        __os = OSEnum.OSX
    elif plt in ["win32", "cygwin"]:
        __os = OSEnum.WIN

    return __os



"""
Pepsin Main CLI Program, the cli entrypoint
"""
import os
import sys
from typing import List, Optional

from pepsin.base_io import IOBase
from pepsin.const import COMMAND_NOT_FOUND_ERROR
from pepsin.utils.base import get_commands, load_command_class
from pepsin.version import get_version


class CLI(IOBase):
    """
    CLI Class that handles cli arguments, parses them
    and runs command through execute method
    """

    def __init__(self, argv=None):
        super().__init__()
        self.argv: List[str] = sys.argv[:] if not argv else argv
        self.program_name = os.path.basename(self.argv[0])

        if self.program_name == "__main__.py":
            self.program_name = "python -m pepsin"

        try:
            self.command = self.argv[1]
        except IndexError:
            self.command = "help"

    def get_arg(self, index: int) -> Optional[str]:
        """
        Gets argument with index, if index out of range
        then return None
        Args:
            index: int | List index

        Returns: argument string or none

        """
        try:
            return self.argv[index]
        except IndexError:
            return None

    def help_text(self) -> str:
        """
        Gets all available commands and format them to build help text
        which contains the name of the command and short description of it
        Returns: str | List of commands with short description

        """
        version = get_version()
        if self.command in ["--help", "-h", "help"]:
            string = [
                f"pepsin v{version}",
                "Type the name of the command and --help for help on a"
                " specific command",
                "",
                "Available commands: ",
            ]
            for command in get_commands():
                command_class = load_command_class(command)
                string.append(f"{command} | {command_class.short_description}")

            return "\n".join(string)

        command_class = load_command_class(self.command)
        return command_class.format_help(self.command)

    def execute(self):
        """
        CLI Execute method / Main execution method
        Returns: None

        """
        if self.command in ["--help", "-h", "help"]:
            help_text = self.help_text()
            self.output(help_text)

        elif self.command == "--version" or self.argv[1:] == "--version":
            version = get_version()
            self.output(version + "\n")

        else:
            try:
                command_class = load_command_class(self.command)
                command_class.run(self.argv)
            except ModuleNotFoundError:
                self.error(f"`{self.command}` Command not available")
                self.error(COMMAND_NOT_FOUND_ERROR)


def main():
    """
    Driver function
    Returns:

    """
    runner = CLI()
    runner.execute()



"""
Template handling module
"""
import os
from typing import List

from pepsin.const import TEMPLATE_DIR
from pepsin.error import TemplateDoesNotExistError


class Template:
    """
    Template class to save and format Template Files
    """

    def __init__(self, template_name, save_as=None, context=None):
        self.template_name = template_name
        self.file = ""
        self.save_as = template_name if not save_as else save_as
        self.context = {} if not context else context

    def read(self):
        """
        Reads a given template file and saves template file data
        to self.file
        Returns:

        """
        temp_name = TEMPLATE_DIR / self.template_name

        if not os.path.isfile(temp_name):
            # Throw if templates does not exist
            raise TemplateDoesNotExistError(
                f"Template {temp_name} does not exist"
            )

        with open(temp_name, "r", encoding="utf-8") as file:
            self.file = file.read()

    def format(self):
        """
        Reads the file, stores data in the file attribute
        Then exchanges the context data, that starts with `$`
        For example in a templates file a context variable named
        `file`, in the templates that must be written as following
        `$file`
        While initializing `Template` class along with context parameter
        Template("templates.py", context={"file": "FooFile"})
        'save' method will save the file where $file will be replaced with FooFile

        Returns:

        """
        for key, val in self.context.items():
            self.file.replace(f"${key}", val)

    def save(self):
        """
        Saves templates in the working directory
        Returns:
        """
        self.read()
        self.format()
        working_dir = os.getcwd()
        if not os.path.isfile(os.path.join(working_dir, self.save_as)):
            with open(
                f"{working_dir}/{self.save_as}", "w", encoding="utf-8"
            ) as file:
                file.write(self.file)


class TemplateList:
    """
    Template List DataStructure
    """

    def __init__(self):
        self.templates: List[Template] = []

    def add_template(self, *args):
        """
        Adds templates in the queue
        Args:
            *args: List[Templates] list of templates

        Returns:

        """
        self.templates += args

    def save(self):
        """
        Executes the template.save() method for
        all the given templates
        Returns:

        """
        for template in self.templates:
            template.save()



from setuptools import find_packages, setup

from pepsin import version

with open("./README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def read_requirements():
    with open("./requirements.txt", "r") as req:
        content = req.read()
        requirements = content.split("\n")
    return requirements


VERSION = version.get_version()

setup(
    name="pepsin",
    version=VERSION,
    include_package_data=True,
    description="Python PIP Toolchain cli",
    author="Khan Asfi Reza",
    author_email="khanasfireza10@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests*",)),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=read_requirements(),
    package_data={
        "": [
            "pepsin/templates",
            "pepsin/templates/*",
        ]
    },
    entry_points="""
        [console_scripts]
        pepsin=pepsin.main:main
    """,
    python_requires=">=3.6",
)



from typing import Union

import pytest

from pepsin.base import Base
from pepsin.error import InvalidCommandError


class CommandTest(Base):
    def execute(self, *args, **kwargs) -> Union[None, str]:
        raise InvalidCommandError("Test Error")


@pytest.fixture
def test_base():
    return CommandTest()


def test_format_help(test_base):
    string = test_base.format_help("Test Command")
    assert "Test Command" in string


def mock_exit(code):
    return


def test_error(test_base, monkeypatch):
    monkeypatch.setattr("sys.exit", mock_exit)
    try:
        test_base.run(argv=["pepsin", "command"])
    except InvalidCommandError as e:
        assert e.return_code == 1



from test.utils import command_path, set_subprocess
from urllib.error import URLError

import pytest

from pepsin.commands.upgrade import Command
from pepsin.config import PepsinConfig
from pepsin.utils import write_file


@pytest.fixture
def upgrade_command():
    return Command()


def test_install(upgrade_command):
    upgrade_command.run(
        ["pepsin", "upgrade", "flask", "pip", "-r", "requirements.txt"]
    )
    conf = PepsinConfig()
    configs = conf.format_config()
    assert configs.get("libraries") == ["flask"]


def test_install_with_req(upgrade_command):
    write_file("requirements.txt", "\n".join(["django", "flask"]))
    upgrade_command.run(["pepsin", "upgrade", "-r", "requirements.txt"])
    conf = PepsinConfig()
    assert conf.libraries == ["django", "flask"]


def test_install_without_req(upgrade_command):
    conf = PepsinConfig()
    conf.update(libraries=["django", "flask"])
    upgrade_command.run(["pepsin", "upgrade"])
    assert conf.libraries == ["django", "flask"]


def test_upgrade_pip_fail(upgrade_command, monkeypatch, capsys):
    def raise_urlopen_error(*args, **kwargs):
        raise URLError("TEST")

    monkeypatch.setattr("urllib.request.urlopen", raise_urlopen_error)
    upgrade_command.run(["pepsin", "upgrade", "pip"])
    assert "Unable to upgrade" in capsys.readouterr().err



import pytest

from pepsin.base_io import Input, PromptHandler


@pytest.fixture
def prompt_handler():
    return PromptHandler()


def test_prompt_handler(prompt_handler, monkeypatch):
    """
    Test prompt handler
    """
    prompt_handler.add_prompt(
        Input(name="name", type=str, title="title", required=False)
    )
    assert prompt_handler.__len__() == 1
    # Test 2
    assert not prompt_handler.is_prompt_complete()
    prompt_handler.add_prompts(
        Input(name="title", type=str, title="title", required=False),
        Input(name="age", type=int, title="age", required=False),
    )
    # Test 3
    assert prompt_handler.__len__() == 3
    monkeypatch.setattr("builtins.input", lambda _: "Test")
    prompt_handler.prompt()
    # Test 4
    assert prompt_handler.is_prompt_complete()
    assert prompt_handler.answers.name == "Test"



"""
Upgrade alias
"""
from pepsin.commands.upgrade import Command as UpgradeCommand


class Command(UpgradeCommand):
    """
    Update command / Upgrade Command
    """

    short_description = "Update or upgrade library"
    help = """Upgrades a particular or multiple libraries
        `$pepsin update <library>`
        Example: `$pepsin update django`
        """



"""
Constants that will be used throughout the project
"""
from pathlib import Path

COMMAND_NOT_FOUND_ERROR = "use 'pepsin help' to find commands"
PEPSIN_ROOT = Path(__file__).resolve().parent
TEMPLATE_DIR = PEPSIN_ROOT / "templates"
PIP_DL_LINK = "https://bootstrap.pypa.io/get-pip.py"



"""
Uninstall a library
"""
from argparse import ArgumentParser

from pepsin.base import Base
from pepsin.config import PepsinConfig
from pepsin.pyhandler import PyHandler
from pepsin.utils import get_default


class Command(Base):
    """
    Handles library uninstallation
    """

    short_description = "Uninstall library"
    help = """Uninstall a particular or multiple libraries
    `$pepsin uninstall <library>`
    Example: `$pepsin uninstall falcon`
    """

    def add_argument(self, parser: ArgumentParser):
        parser.add_argument(
            "libraries_to_uninstall",
            nargs="+",
            metavar="libs",
            type=str,
            help="Libraries to uninstall",
            default="",
        )

    def execute(self):
        libs = self.command_data.get("libraries_to_uninstall")
        libs = get_default(libs, [])
        if libs:
            config = PepsinConfig()
            py_handler = PyHandler(pepsin_config=config)
            passed = py_handler.uninstall_libraries(libs)[0]
            config.remove_libraries(passed)



import os
from pathlib import Path

import pytest
import yaml

from pepsin.yml import YAMLConfig, dict_to_yaml, yaml_to_dict


@pytest.fixture
def path():
    path = Path(__file__).resolve().parent / "data"
    os.chdir(path=path)
    return path


def test_yaml_load():
    yaml_data = """
    data: Test
    name: pepsin
    """
    data = yaml.load(yaml_data, yaml.Loader)
    assert data["name"] == "pepsin"


def test_load_yaml_data(path):
    data = yaml_to_dict("test.yaml")
    assert data.get("name") == "pepsin"


def test_load_dict_to_yaml(path):
    __dict = {"name": "pepsin"}
    dict_to_yaml(__dict, "pytest.yaml")
    data = yaml_to_dict("pytest.yaml")
    assert data.get("name") == "pepsin"
    os.remove("pytest.yaml")


def test_yaml_config_loader(path):
    Config = YAMLConfig("test.yaml")
    assert Config.get("name") == "pepsin"


def test_dict_to_yaml_config_loader(path):
    conf = YAMLConfig("pytest.yaml")
    # Test append method
    conf.append(name="pepsin")
    conf.append(number=12)
    assert conf.get("name") == "pepsin"
    # Test remove method
    conf.remove("number")
    assert conf.get("number") is None
    conf.save()
    # Test Save method
    assert conf.get("name") == "pepsin"
    # Test get_filename
    filename = conf.get_filename()
    assert filename == "pytest.yaml"
    # Test 2 function mem location
    _con = conf.read_from_yaml()
    _copy = conf.get_config()
    assert id(_con) != id(_copy)
    os.remove("pytest.yaml")


def test_empty_yaml(path):
    __conf = YAMLConfig("temp_test.yaml")
    assert __conf.get_filename() == "temp_test.yaml"
    __conf.read_from_yaml()
    assert len(__conf.get_config()) == 0
    assert __conf.get_filename() == "temp_test.yaml"



"""
A module for input output handling,
_TODO: Use Rich Library
"""
import abc
import dataclasses
import sys
from collections import namedtuple
from io import TextIOBase
from typing import List, Union

from colorama import Fore, init


class OutputWrapper(TextIOBase):
    """
    Output wrapper that uses std.out and enforces color
    """

    color_map = {
        "success": Fore.GREEN,
        "error": Fore.RED,
        "warning": Fore.YELLOW,
        "primary": Fore.CYAN,
    }

    def __init__(self, out: TextIOBase, ending: str = "\n"):
        super().__init__()
        init(autoreset=True)
        self._out = out
        self.ending = ending

    def flush(self):
        """
        Flush out text
        Returns:

        """
        if hasattr(self._out, "flush"):
            self._out.flush()

    def write(
        self,
        msg: str = "",
        ending: str = None,
        msg_type: str = "text",
        enforce_color: str = "",
    ):
        """
        Outputs text
        Args:
            msg: str | Msg that will be in the output
            ending: str | Ending of the msg
            msg_type: str | Options - success, error, warning, primary
            enforce_color: str | Color of the text

        Returns:

        """
        ending = self.ending if ending is None else ending
        if ending and not msg.endswith(ending):
            msg += ending
        color = self.color_map.get(msg_type, "")
        color = enforce_color if enforce_color else color
        self._out.write(color + msg)


class IOBase:
    """
    IOBase is a superclass for CLI Output and Error Handling
    """

    def __init__(self):
        self.stdout = OutputWrapper(sys.stdout)
        self.stderr = OutputWrapper(sys.stderr)

    def output(self, text):
        """
        Text output on cli
        """
        self.stdout.write(text)

    def output_flush(self):
        """
        Output cli flush
        """
        self.stdout.flush()

    def error(self, text):
        """
        CLI Output for errors
        """
        self.stderr.write(text)

    def error_flush(self):
        """
        CLI Error flush
        """
        self.stderr.flush()


@dataclasses.dataclass
class CLIIOBase(abc.ABC):
    """
    CLI Input or output base
    """

    name: str
    title: str
    skip: bool = False

    @abc.abstractmethod
    def prompt(self) -> Union[bool, None, int, str]:
        """
        Read a string from standard input. Then input data
        is fixed and converted to certain data type given by
        `Input.type` or Prints out statement
        """

    @abc.abstractmethod
    def prompt_as_dict(self):
        """
        Returns prompt data dictionary
        """


@dataclasses.dataclass
class Output(IOBase, CLIIOBase):
    """
    Outputs a title while scanning input from the cli
    """

    color: str = ""

    def __post_init__(self):
        self.output = OutputWrapper(sys.stdout)

    def prompt_as_dict(self):
        """
        Returns empty dictionary
        """
        self.prompt()
        return {}

    def prompt(self):
        """
        Outputs title in the cli
        """
        self.output.write(self.title, enforce_color=self.color)


@dataclasses.dataclass
class Input(CLIIOBase):
    """
    params:
        name: str | Name of the input
        title: str | Title that will be displayed in the cli while input prompt
        type: type | Type of input, if options exist then type will be integer for default
        default: Any | Default value of the input
        required: bool | Is the input required or skip-able
        options: list | Available option list
        skip: bool | Skips if argument is given
    """

    type: type = str
    default: Union[str, int, bool, None] = None
    required: bool = True
    options: list = dataclasses.field(default_factory=list)

    def __post_init__(self):
        if self.options:
            self.type = int

    def input_title(self) -> str:
        """
        This method returns The prompt string,
        that will be printed along with input
        """
        title = self.title
        if self.type is bool:
            title += " (y/n)"

        if self.options:
            title += "\n"
            for key, option in enumerate(self.options):
                title += f"{key + 1}. {option}\n"

        default = "" if not self.default else self.default

        title += f"[{default}]:"
        return title

    def prompt(self) -> Union[bool, None, int, str]:
        """
        Read a string from standard input. Then input data
        is fixed and converted to certain data type given by
        `Input.type`
        """
        while True:
            title = self.input_title()
            inp = input(title)
            __final = self.default
            __final = inp if inp else __final

            if self.type is int and inp.isdigit():
                data = int(inp)
                if self.options and 0 <= data <= len(self.options):
                    # Option to select and return
                    __final = self.options[data - 1]
                elif not self.options:
                    __final = data
                else:
                    continue

            return inp.lower() in ["1", "y"] if self.type is bool else __final

    def prompt_as_dict(self):
        """
        Returns input in dictionary format
        """
        return {self.name: self.prompt()}


class PromptHandler:
    """
    Handles group of input together and stores in the object in
    multiple form
    """

    def __init__(self, *inputs, options=None):
        self.__inputs: List[Input] = []
        self.options = options
        self.add_prompts(*inputs, options=self.options)
        self.__answers = {}

    def update_options(self, options):
        """
        Updates Argument option

        Returns:

        """
        self.options = options

    def __len__(self):
        return len(self.__inputs)

    def is_prompt_complete(self):
        """
        Checks if input has been taken from cli
        """
        return len(self.__answers) > 0

    def add_prompt(self, _input: Input, options=None):
        """
        Add one input in the prompt list, that will be
        executed serially
        _input: `Input` Object
        """
        if not options:
            options = self.options
        if options and options.get(_input.name, None) and _input.skip:
            return
        self.__inputs.append(_input)

    def add_prompts(self, *inputs, options=None):
        """
        Add multiple input in the prompt list, that will be
        executed serially
        inputs: List of `Input` Object
        """
        if not options:
            options = self.options
        self.__inputs += [
            inp
            for inp in inputs
            if not (options and options.get(inp.name, None) and inp.skip)
        ]

    def get_answers(self):
        """
        Returns answers in dict format
        """
        return self.__answers

    def get(self, key):
        """
        Returns one answer using key
        """
        return self.__answers.get(key, None)

    def prompt(self, options=None):
        """
        Get input from all the given Input Object
        returns: Dictionary of answers
        """
        for inp in self.__inputs:
            if not (inp.skip and options and options.get(inp.name, None)):
                self.__answers.update(inp.prompt_as_dict())
        return self.__answers

    @property
    def answers(self):
        """
        Returns answer in Class Object Form
        """
        return namedtuple("Answer", self.__answers.keys())(
            *self.__answers.values()
        )
